import numpy as np
from scipy import signal, stats
from sklearn.decomposition import FastICA
from pywt import wavedec
from pywt import waverec
import matplotlib.pyplot as plt
from matplotlib import colors as mcolors
from matplotlib.collections import LineCollection


def blinkremove(eeg_signal, srate, chnum, debug_mode=False):
    """
    removes blink artifact from signal
    :param eeg_signal: input eeg signal
    :param srate: sampling rate
    :return: blink removed signal
    """

    ############### ICA ######################
    # the length of samples must be even
    del_flag = 0
    if (np.shape(eeg_signal)[1]) % 2 != 0:
        eeg_signal = np.delete(eeg_signal, 0, 1)
        del_flag = 1

    # chnum = eeg_signal.shape[0]
    eeg_signal = eeg_signal.transpose()
    ica = FastICA(n_components=chnum, whiten=True)

    eeg_unmixed = ica.fit_transform(eeg_signal)  # Reconstruct signals
    iteres = ica.n_iter_
    print("iteration needed:", iteres)
    if iteres > 199:
        return 'converge error'
    myvar = np.sqrt(np.var(eeg_unmixed))
    eeg_unmixed = eeg_unmixed / myvar  # this is to handle the scikit learn bug!
    A_ = ica.mixing_  # Get estimated mixing matrix
    eeg_unmixed = eeg_unmixed.transpose()
    if debug_mode:
        eegplot(eeg_unmixed, srate, 'ica unmixed signal')

    ############### Kurtosis ######################
    Kurtosis = stats.kurtosis(eeg_unmixed, axis=-1)

    # thresholding
    lessKurt = Kurtosis.copy()
    # # throw maximum away and measure std with others
    # maxIndex = np.where(lessKurt == max(lessKurt))
    # maxIndex = maxIndex[0][0]
    # lessKurt = np.array(lessKurt, maxIndex)
    mKurt = np.mean(lessKurt)
    sdKurt = np.std(lessKurt)
    CI = 0.90  # two tail 0.9
    tN_1 = stats.t.ppf(1 - (1 - CI) / 2, chnum - 1)
    kuUpperLim = mKurt + sdKurt / np.sqrt(chnum) * tN_1

    if debug_mode:
        plt.figure('Kurtosis of ICs')
        plt.bar(list(range(1, chnum + 1)), Kurtosis)
        plt.title('Kurtosis of ICs')
        plt.plot(list(range(0, chnum + 2)), kuUpperLim * np.ones(chnum + 2), 'r', 'LineWidth', 1)

    ############### Skewness ######################
    skew = stats.skew(eeg_unmixed, axis=-1)
    # thresholding
    lessskew = skew.copy()
    mskew = np.mean(lessskew)
    sdskew = np.std(lessskew)
    CI = 0.95  # two tail 0.95
    tN_1 = stats.t.ppf(1 - (1 - CI) / 2, chnum - 1)
    skUpperLim = mskew + sdskew / np.sqrt(chnum) * tN_1
    skLowerLim = mskew - sdskew / np.sqrt(chnum) * tN_1

    if debug_mode:
        plt.figure('Skweness of ICs')
        plt.bar(list(range(1, chnum + 1)), skew)
        plt.title('Skweness of ICs')

        plt.plot(list(range(0, chnum + 2)), skUpperLim * np.ones(chnum + 2), 'r', 'LineWidth', 1)
        plt.plot(list(range(0, chnum + 2)), skLowerLim * np.ones(chnum + 2), 'r', 'LineWidth', 1)

    ############### Perform Wavelet on Blink Components ######################
    isblink = (skew > skUpperLim) | (skew < skLowerLim) | (Kurtosis > kuUpperLim)

    eeg_ica_noblink = eeg_unmixed.copy()

    # for each blink detected IC
    for xcomp in range(chnum):
        if isblink[xcomp]:
            blink = eeg_unmixed[xcomp, :]
            coeffs = wavedec(data=blink, level=8, wavelet='bior4.4')

            # null the blink wavelet coefficients
            newcoeffs = coeffs.copy()
            K = np.zeros(9)
            for ix in range(9):
                temp = coeffs[ix]
                sigma2 = np.median(abs(temp) / 0.6745)
                K[ix] = np.sqrt(2 * np.log10(len(temp)) * sigma2)
                # type I
                Indx = list(np.where(np.abs(temp) > K[ix])[0])
                temp[Indx] = 0
                newcoeffs[ix] = temp

            eeg_ica_noblink[xcomp] = waverec(newcoeffs, 'bior4.4')

    eeg_ica_noblink = np.array(eeg_ica_noblink)
    if debug_mode:
        eegplot(eeg_ica_noblink, srate, 'ICA Blink Removed')

    ############### Reconstruct EEG signals #######################

    # inverse of ICA
    eeg_ica_noblink = eeg_ica_noblink * myvar  # this is to handle the scikit learn bug!
    eeg_ica_noblink = eeg_ica_noblink.transpose()
    eeg_blink_removed = ica.inverse_transform(eeg_ica_noblink)
    eeg_blink_removed = eeg_blink_removed.transpose()
    eeg_signal = eeg_signal.transpose()
    # Spectral Coherence Measure
    if debug_mode:
        cxy = []
        for xch in range(chnum):
            # [cxy(xChannel,:), fc] = mscohere(data(xChannel,:), FinalEEG(xChannel,:), hamming(512), 256, 256, 2000)
            f, Cxy = signal.coherence(eeg_signal[xch], eeg_blink_removed[xch], srate, nperseg=1024)
            cxy.append(Cxy)

        mcxy = np.mean(np.array(cxy), axis=0)
        plt.figure("Spectral Coherence Measurement")
        plt.stem(f, mcxy)

    if del_flag == 1:
        eeg_blink_removed = np.hstack((eeg_blink_removed, np.tile(eeg_blink_removed[:, [-1]], 1)))
    return eeg_blink_removed


def eegplot(eeg_data, srate, title, fig=None, block=False):
    """
    the function to plot eeg data
    :param eeg_data: get eeg data
    :param srate: sampling rate of eeg signal
    :param title: string of plot title
    :return: plot the data
    """
    eeg_data = np.array(eeg_data)
    eeg_data[:-1, :] = signal.detrend(eeg_data[:-1, :], axis=-1, type='linear')
    n_samples, n_rows = np.shape(eeg_data)[1], np.shape(eeg_data)[0]

    # normalize each column to be able to show
    for i in range(n_rows):
        eeg_data[i, :] = eeg_data[i, :] - np.mean(eeg_data[i, :])
        eeg_data[i, :] = eeg_data[i, :] / (np.std(eeg_data[i, :]) + 1)

    if fig is None:
        fig = plt.figure("EEG plot of {}".format(title))

    axes = plt.axes()
    # Load the EEG data
    data = np.transpose(eeg_data)
    t = np.arange(n_samples) / srate

    # Plot the EEG
    ticklocs = []
    axes.set_xlim(0, t.max())
    # ax2.set_xticks(np.arange(10))

    segs = []
    y1 = 0
    for i in range(n_rows):
        dmin = data[:, i].min()
        dmax = data[:, i].max()
        dr = (dmax - dmin)
        segs.append(np.column_stack((t, data[:, i])))
        ticklocs.append(y1)
        y1 = y1 + dr

    y0 = data[:, 0].min()
    axes.set_ylim(y0, y1)

    offsets = np.zeros((n_rows, 2), dtype=float)
    offsets[:, 1] = ticklocs

    colors = [mcolors.to_rgba(c)
              for c in plt.rcParams['axes.prop_cycle'].by_key()['color']]
    lines = LineCollection(segs, offsets=offsets, transOffset=None, linewidths=0.5,
                           colors=colors, linestyle='solid')
    axes.add_collection(lines)

    # Set the yticks to use axes coordinates on the y axis
    axes.set_yticks(ticklocs)

    ch_name_list = []
    for i in range(n_rows):
        ch_name_list.append('CH '+str(i))
    axes.set_yticklabels(ch_name_list)

    axes.set_xlabel('Time (s)')
    plt.suptitle(title)

    plt.tight_layout()
    plt.show()
